%{
#include "y.tab.h"
#include <stdlib.h>

void yyerror(char *);
%}

%%

0x[0-9a-fA-F]+ {
        yylval.i = strtoll(yytext+2, NULL, 16);
        return INTEGER;
}

0b[01]+ {
        yylval.i = strtoll(yytext+2, NULL, 2);
        return INTEGER;
}

-?[0-9]+\.[0-9]+ {
        yylval.d = strtof(yytext, NULL);
        return DOUBLE;
}

[0-9]+ {
        yylval.i = strtoll(yytext, NULL, 10);
        return INTEGER;
}

\*\*         { return POWER; }
\<\<         { return SHIFT_LEFT; }
>>           { return SHIFT_RIGHT; }
>>s          { return SHIFT_RIGHT_SIGNED; }
rol8         { return ROL8; }
rol16        { return ROL16; }
rol32        { return ROL32; }
rol64        { return ROL64; }
ror8         { return ROR8; }
ror16        { return ROR16; }
ror32        { return ROR32; }
ror64        { return ROR64; }
slb          { return SLB; }
modinv32     { return MODINV32; }
modinv64     { return MODINV64; }
dump_ieee754 { return DUMP_IEEE754; }
as_float     { return AS_FLOAT; }
as_double    { return AS_DOUBLE; }
log2         { return LOG2; }
[-+*/^\n~(),<>s|&%] {
        return *yytext;
}

[ \t]       ;       /* skip whitespace */

.           printf("(flex) unknown character: %c\n", yytext[0]);

%%

int yywrap(void)
{
        return 1;
}

void flex_set_str(char *s)
{
        YY_BUFFER_STATE buf;
        buf=yy_scan_string (s);
        yy_switch_to_buffer (buf);
};

void flex_cleanup()
{
        yy_delete_buffer( YY_CURRENT_BUFFER );
};

/* vim: set expandtab ts=8 sw=8 : */

