%{
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <assert.h>
#include <x86intrin.h>

#include <readline/readline.h>
#include <readline/history.h>

int yylex(void);
void yyerror(char *);

bool IsInteger (double d)
{
        return floor(d)==d;
};

// copypasted from Hacker's Delight, Chapter 8, p.155
// w=u*v, m=number of 16-bit words in u[], n=... in v[]
void mulmns(uint16_t w[], uint16_t u[], uint16_t v[], int m, int n)
{
        unsigned int k, t, b;
        int i, j;
        for (i = 0; i < m; i++)
                w[i] = 0;
        for (j = 0; j < n; j++)
        {
                k = 0;
                for (i = 0; i < m; i++)
                {
                        t = u[i]*v[j] + w[i + j] + k;
                        w[i + j] = t; // (I.e., t & 0xFFFF).
                        k = t >> 16;
                }
                w[j + m] = k;
        }
        // Now w[] has the unsigned product.
        // sign of w[] is not corrected
};

#define SHORTS_PER_UINT32 sizeof(uint32_t)/sizeof(uint16_t)
#define SHORTS_PER_UINT64 sizeof(uint64_t)/sizeof(uint16_t)

// reworked from https://github.com/gcc-mirror/gcc/blob/1cb6c2eb3b8361d850be8e8270c597270a1a7967/gcc/loop-iv.c#L1291
// this function solves the equation:
// a * x = 1 + b * 2^32
// ... and returns a, b
// see also: https://en.wikipedia.org/wiki/B%C3%A9zout's_identity
bool modinv32 (uint32_t x, uint32_t *a, uint32_t *b)
{
        uint32_t input_x=x, rslt=1;

        for (int i=0; i<31; i++)
        {
                rslt = (rslt * x);
                x = (x * x);
        }

        // get the second coefficient of Bézout's identity:
        uint32_t tmp[2];
        mulmns((uint16_t*)tmp, (uint16_t*)&rslt, (uint16_t*)&input_x, SHORTS_PER_UINT32, SHORTS_PER_UINT32);

        if (tmp[0]!=1)
                return false; // not coprimes

        *a=rslt;
        *b=tmp[1];

        return true;
}

// a * x = 1 + b * 2^64
bool modinv64 (uint64_t x, uint64_t *a, uint64_t *b)
{
        uint64_t input_x=x, rslt=1;

        for (int i=0; i<63; i++)
        {
                rslt = (rslt * x);
                x = (x * x);
        }

        // get the second coefficient of Bézout's identity:
        uint64_t tmp[2];
        mulmns((uint16_t*)tmp, (uint16_t*)&rslt, (uint16_t*)&input_x, SHORTS_PER_UINT64, SHORTS_PER_UINT64);

        if (tmp[0]!=1)
                return false; // not coprimes

        *a=rslt;
        *b=tmp[1];

        return true;
}

void print_bin (uint64_t v)
{
        printf ("0b");

        if (v==0)
        {
                // just print zero
                printf ("0");
                return;
        };

        bool one_printed=false;
        for (int i=64-1; i>=0; i--)
        {
                uint64_t digit=(v>>i) & 1UL;
                if (one_printed)
                        printf ("%"PRIu64, digit);
                else
                        if (digit)
                        {
                                printf ("1");
                                one_printed=true;
                        }
                        else
                        {
                                // supress leading zero
                        };
        };
};

uint8_t rotr8(uint8_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rorb %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotr8(x,r);
#endif
}

uint8_t rotl8(uint8_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rolb %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotl8(x,r);
#endif
}

uint16_t rotr16(uint16_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rorw %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotr16(x,r);
#endif
}

uint16_t rotl16(uint16_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rolw %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotl16(x,r);
#endif
}

uint32_t rotr32(uint32_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rorl %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotr(x,r);
#endif
}

uint32_t rotl32(uint32_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("roll %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotl(x,r);
#endif
}

#if 0
uint64_t rotr64(uint64_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rorq %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotr64(x,r);
#endif
}

uint64_t rotl64(uint64_t x, uint8_t r)
{
#ifdef __GNUC__
        __asm("rolq %1,%0" : "+r" (x) : "c" (r));
        return x;
#endif
#ifdef _MSC_VER
        return _rotl64(x,r);
#endif
}
#endif

uint64_t rotr64 ( uint64_t x, uint8_t r )
{
        return (x >> r) | (x << (64 - r));
}

uint64_t rotl64 ( uint64_t x, uint8_t r )
{
        return (x << r) | (x >> (64 - r));
}

// http://stackoverflow.com/questions/11376288/fast-computing-of-log2-for-64-bit-integers
static const int tab64[64]=
{
        63,  0, 58,  1, 59, 47, 53,  2,
        60, 39, 48, 27, 54, 33, 42,  3,
        61, 51, 37, 40, 49, 18, 28, 20,
        55, 30, 34, 11, 43, 14, 22,  4,
        62, 57, 46, 52, 38, 26, 32, 41,
        50, 36, 17, 19, 29, 10, 13, 21,
        56, 45, 25, 31, 35, 16,  9, 12,
        44, 24, 15,  8, 23,  7,  6,  5
};

uint64_t uint64_log2 (uint64_t value)
{
        value |= value >> 1;
        value |= value >> 2;
        value |= value >> 4;
        value |= value >> 8;
        value |= value >> 16;
        value |= value >> 32;
        return tab64[((uint64_t)((value - (value >> 1))*0x07EDD5E59A4E28C2)) >> 58];
}

// http://stackoverflow.com/questions/101439/the-most-efficient-way-to-implement-an-integer-based-power-function-powint-int
// TODO check overflow
uint64_t ipow(uint64_t base, uint64_t exp)
{
        uint64_t result = 1;
        while (exp)
        {
                if (exp & 1)
                        result *= base;
                exp >>= 1;
                base *= base;
        }

        return result;
}

char* str_trim_all_lf_cr_right (char *in)
{
        size_t slen=strlen(in);
        if (slen==0)
                return in;
        int last_char_pos=slen-1;

        if (in[last_char_pos]=='\n' || in[last_char_pos]=='\r')
        {
                in[last_char_pos]=0;
                str_trim_all_lf_cr_right(in);
        };

        return in;
};

// -----------------------
uint64_t last_result;
bool silent;
int cnt=1;
#define RESULTS_TOTAL 1024
uint64_t results[RESULTS_TOTAL];

void print_val(uint64_t a)
{
        last_result=a;
        results[cnt]=a;

        if (silent)
                return;
        printf("[%d] (unsigned) %" PRIu64 " 0x%" PRIx64 " ", cnt, a, a);
        print_bin (a);
        
        // ASCII?
        if (a>=0x20 && a<=0x7f)
                printf (" ASCII: '%c'", (char)a);
        
        printf ("\n");

        int64_t s=*(int64_t*)&a;
        if (s<0)
                printf("[%d] (signed) -%" PRIu64 "\n", cnt, -s);
};

// set leading bits (to 1)
uint64_t slb (uint64_t x)
{
        uint64_t rt=x;
        for (int i=64-1; i>=0; i--)
        {
                if ((x>>i) & 1ULL)
                        break;
                rt|=1ULL<<i;
        };
        return rt;
};

uint64_t modinv (uint64_t in, int mod)
{
        if (silent==false)
        {
                double result=pow (2, (double)mod) / (double)in;
                if (IsInteger(result)==false)
                        printf ("Warning, result is not integer: 2^%d/input=%f\n", mod, result);
                else
                        printf ("2^%d/input=%f\n", mod, result);
        };

        if (silent==false)
                printf ("Solving the equation %"PRIu64"x = 1+y*2^%d\n", in, mod);
        if (mod==32)
        {
                uint32_t x, y;
                if (modinv32(in, &x, &y)==false)
                {
                        if (silent==false)
                                printf ("No solutions\n");
                        return 0;
                }
                else
                {
                        if (silent==false)
                                printf ("x=%u, y=%d\n", x,y);
                        return x;
                };
        }
        else if (mod==64)
        {
                uint64_t x, y;
                if (modinv64(in, &x, &y)==false)
                {
                        if (silent==false)
                                printf ("No solutions\n");
                        return 0;
                }
                else
                {
                        if (silent==false)
                                printf ("x=%"PRIu64", y=%d\n", x,y);
                        return x;
                };
        }
        
        assert(0);
};

struct float_as_struct
{
        unsigned int fraction : 23;
        unsigned int exponent : 8;
        unsigned int sign : 1;
};

struct double_as_struct
{
        uint64_t fraction : 52;
        int exponent : 11;
        int sign : 1;
};

void dump_ieee754 (double i)
{
        uint64_t dtmp;
        struct double_as_struct dtmps;

        float f=i;
        uint32_t stmp;
        struct float_as_struct stmps;

        memcpy (&dtmp, &i, sizeof(double));
        memcpy (&dtmps, &i, sizeof(double));
        
        memcpy (&stmp, &f, sizeof(float));
        memcpy (&stmps, &f, sizeof(float));

        printf ("as 64-bit double:\n");
        printf ("0x%" PRIx64 "\n", dtmp);
        printf ("sign: %d exponent: 0x%x fraction:0x%" PRIx64 "\n", dtmps.sign, dtmps.exponent, dtmps.fraction);
        printf ("as 32-bit float:\n");
        printf ("0x%x\n", stmp);
        printf ("sign: %d exponent: 0x%x fraction:0x%x\n", stmps.sign, stmps.exponent, stmps.fraction);
        
};

void as_float(uint64_t x)
{
        uint32_t it=x;
        float ft;
        memcpy (&ft, &it, sizeof(uint32_t));
        printf ("%f\n", ft);
};

void as_double(uint64_t x)
{
        double dt;
        memcpy (&dt, &x, sizeof(uint64_t));
        printf ("%f\n", dt);
};

uint64_t division (uint64_t a, uint64_t b)
{
        double result=(double)a / (double)b;
        if (IsInteger(result)==false && silent==false)
                printf ("Warning, result is not integer: %f; remainder: %" PRIu64 " %% %" PRIu64 " = %" PRIu64 "\n", result, a, b, a % b);
        
        return a/b;
};

%}

%union 
{
        uint64_t i;
        double d;
}

%token INTEGER DOUBLE ROL8 ROL16 ROL32 ROL64 ROR8 ROR16 ROR32 ROR64 SLB MODINV32 MODINV64 DUMP_IEEE754 AS_FLOAT AS_DOUBLE SHIFT_LEFT SHIFT_RIGHT SHIFT_RIGHT_SIGNED POWER LOG2

%left '|'
%left '^'
%left '&'
%left SHIFT_LEFT SHIFT_RIGHT SHIFT_RIGHT_SIGNED
%left '-' '+'
%left '*' '/' '%'
%left POWER

%type <i> expr
%type <d> DOUBLE
%type <i> INTEGER

%%

input:
        expr                            { print_val($1); }     // these are 'procedures' (which doesn't return anything)
        | DUMP_IEEE754 '(' DOUBLE ')'   { dump_ieee754 ($3); }
        | DUMP_IEEE754 '(' INTEGER ')'  { dump_ieee754 ((double)$3); }
        | AS_FLOAT '(' expr ')'         { as_float ($3); }
        | AS_DOUBLE '(' expr ')'        { as_double ($3); }
        ;

expr:
        INTEGER
        | '%'                           { $$ = last_result; }
        | '%' INTEGER                   { if ($2==0 || $2>cnt) printf ("Warning: no such result (yet?), using 0\n"); $$ = results[$2]; }
        | '(' expr ')'                  { $$ = $2; }
        | ROL8 '(' expr ',' expr ')'    { $$ = ($3 & (~0xFF))  | rotl8 ($3, $5); }
        | ROL16 '(' expr ',' expr ')'   { $$ = ($3 & (~0xFFFF)) | rotl16 ($3, $5); }
        | ROL32 '(' expr ',' expr ')'   { $$ = ($3 & (~0xFFFFFFFF)) | rotl32 ($3, $5); }
        | ROL64 '(' expr ',' expr ')'   { $$ = rotl64 ($3, $5); }
        | ROR8 '(' expr ',' expr ')'    { $$ = ($3 & (~0xFF))  | rotr8 ($3, $5); }
        | ROR16 '(' expr ',' expr ')'   { $$ = ($3 & (~0xFFFF)) | rotr16 ($3, $5); }
        | ROR32 '(' expr ',' expr ')'   { $$ = ($3 & (~0xFFFFFFFF)) | rotr32 ($3, $5); }
        | ROR64 '(' expr ',' expr ')'   { $$ = rotr64 ($3, $5); }
        | MODINV32 '(' expr ')'         { $$ = modinv ($3, 32); }
        | MODINV64 '(' expr ')'         { $$ = modinv ($3, 64); }
        | SLB '(' expr ')'              { $$ = slb ($3); }
        | LOG2 '(' expr ')'             { $$ = uint64_log2 ($3); }
        | '~' expr                      { $$ = ~$2; }
        | '-' expr                      { $$ = -$2; }
        | expr SHIFT_LEFT expr          { $$ = $1 << $3; }
        | expr SHIFT_RIGHT expr         { $$ = $1 >> $3; }
        | expr SHIFT_RIGHT_SIGNED expr  { $$ = (uint64_t)((int64_t)$1 >> (int64_t)$3); }
        | expr POWER expr               { $$ = ipow ($1, $3); }
        | expr '*' expr                 { $$ = $1 * $3; }
        | expr '/' expr                 { $$ = division($1, $3); }
        | expr '%' expr                 { $$ = $1 % $3; }
        | expr '+' expr                 { $$ = $1 + $3; }
        | expr '-' expr                 { $$ = $1 - $3; }
        | expr '^' expr                 { $$ = $1 ^ $3; }
        | expr '&' expr                 { $$ = $1 & $3; }
        | expr '|' expr                 { $$ = $1 | $3; }
        ;

%%

void yyerror(char *s)
{
        fprintf(stderr, "bison error: %s\n", s);
}

char* strdup(const char*);

int parse(char *s)
{
        int r;
        char* buf=strdup(s);
        str_trim_all_lf_cr_right(buf);
        flex_set_str(buf);
        r=yyparse();
        flex_cleanup();
        free(buf);
        cnt=(cnt+1)&(RESULTS_TOTAL-1);
}

void test()
{
        silent=true;
        parse ("(1+2)*3\n");
        assert(last_result==9);
        parse ("1234\n");
        assert(last_result==1234);
        parse ("-1234\n");
        assert(last_result==0xfffffffffffffb2eUL);
        parse ("modinv32(5)\n");
        assert(last_result==0xcccccccd);
        parse ("modinv32(0xcccccccd)\n");
        assert(last_result==5);
        parse ("modinv64(12345)\n");
        assert(last_result==0x4963847d55d4be09);
        parse ("ror64(1,1)\n");
        assert(last_result==0x8000000000000000UL);
        parse ("ror8(0x12345678,3)\n");
        assert(last_result==0x1234560f);
        parse ("slb(0xff78)\n");
        assert(last_result==0xffffffffffffff78UL);
        parse ("~0xff\n");
        assert(last_result==0xffffffffffffff00UL);
        parse ("0xf << 7\n");
        assert(last_result==0x780);
        parse ("0xf000 >> 7\n");
        assert(last_result==0x1e0);
        parse ("slb(0xf000) >>s 4\n");
        assert(last_result==0xffffffffffffff00UL);
        parse ("slb(0xf000) >>s 7\n");
        assert(last_result==0xffffffffffffffe0UL);
        parse ("0xf0 | 1\n");
        assert(last_result==0xf1);
        parse ("0xf0 ^ 0x10\n");
        assert(last_result==0xe0);
        parse ("0xf0 & 0x10\n");
        assert(last_result==0x10);
        parse ("7 % 2\n");
        assert(last_result==1);
        parse ("0x7fffffffffffffff\n");
        assert(last_result==0x7fffffffffffffff);
        parse ("0x1000000000000000\n");
        assert(last_result==0x1000000000000000);
        parse ("1152921504606846976\n");
        assert(last_result==1152921504606846976);
        parse ("0b1000000000000000000000000000000000000000000000000000000000000\n");
        assert(last_result==1152921504606846976);
        parse ("-1+123\n");
        assert(last_result==122);
        parse ("1+2*100\n");
        assert(last_result==201);
        parse ("2*100+1\n");
        assert(last_result==201);
        parse ("2**16\n");
        assert(last_result==0x10000);
        parse ("2**32\n");
        assert(last_result==0x100000000);
        parse ("10**10\n");
        assert(last_result==10000000000);
        parse ("log2(0b1111)\n");
        assert(last_result==3);
        parse ("%+1\n");
        assert(last_result==4);
        silent=false;
};

// GNU readline
// http://stackoverflow.com/questions/29301242/readline-how-to-list-all-autocomplete-matches-on-double-tab


const char *words[] = {"rol8", "rol16", "rol32", "rol64", "ror8", "ror16", "ror32", "ror64",
        "slb", "modinv32", "modinv64", "dump_ieee754", "as_float", "as_double", "log2", NULL};

// Generator function for word completion.
char *my_generator (const char *text, int state)
{
        static int list_index, len;
        const char *name;

        if (!state)
        {
                list_index = 0;
                len = strlen (text);
        }

        while (name = words[list_index])
        {
                list_index++;
                if (strncmp (name, text, len) == 0)
                        return strdup (name);
        }

        // If no names matched, then return NULL.
        return ((char *) NULL);
}

// Custom completion function
static char **my_completion (const char *text, int start, int end)
{
        // This prevents appending space to the end of the matching word
        rl_completion_append_character = '\0';

        char **matches = (char **) NULL;
        if (start == 0)
        {
                matches = rl_completion_matches ((char *) text, &my_generator);
        }
        return matches;
}

void readline_cycle()
{
        char *buf;
        rl_attempted_completion_function = my_completion;
        char prompt[128];

        while (1)
        {
                sprintf (prompt, "[%d] ", cnt);
                buf = readline(prompt);
                if (buf==NULL)
                        break;
                //if (strcmp (buf, "exit") == 0)
                //      break;
                if (buf[0] == '\0')
                {
                        free (buf);
                        continue;
                }
                else
                {
                        parse (buf);
                        add_history (buf);
                }

                free (buf);
                buf = NULL;
        }
        if (buf != NULL) free (buf);
};

int main(int argc, char *argv[])
{
        test();
        printf ("Programmer's calculator <dennis(a)yurichev.com> https://yurichev.com/progcalc (%s)\n", __DATE__);
        if (argc==2)
        {
                parse (argv[1]);
                return 0;
        };

        memset (results, 0, sizeof(results));
        cnt=1;
        readline_cycle();
};

/* vim: set expandtab ts=8 sw=8 : */

